"""
Copyright 2021-2022 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import errno
import os
import shutil


class AuxiliaryFunctions:
	"""
	Auxiliary functions class.

	Methods
	-------
	rm_rf(path: str) -> None:
		Remove all files and directories specified by path.
	symlink_overwrite(destination: str, link: str) -> None:
		Create new or overwrite existing symbolic link.
	atexit_callback() -> None:
		Sentry at exit integration callback.
	"""

	@staticmethod
	def rm_rf(path: str) -> None:
		"""
		Remove all files and directories specified by path.

		Parameters
		----------
		path: str
			path to directory or file

		Returns
		-------
		None
		"""
		if not os.path.exists(path):
			return
		if os.path.isdir(path):
			shutil.rmtree(path)
		else:
			os.unlink(path)

	@staticmethod
	def symlink_overwrite(destination: str, link: str) -> None:
		"""
		Create new or overwrite existing symbolic link.

		Parameters
		----------
		destination: str
			source file for symbolic link
		link: str
			symbolic link name / path

		Returns
		-------
		None

		Exceptions
		----------
		OSError
			raised if an OS error occurs
		"""
		try:
			os.symlink(destination, link)
		except OSError as os_error:
			if os_error.errno == errno.EEXIST:
				os.remove(link)
				os.symlink(destination, link)
			else:
				raise os_error

	@staticmethod
	def atexit_callback() -> None:
		"""
		Sentry at exit integration callback.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
