"""
Copyright 2021-2022 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class UartConfig:
	"""
	UART interface configuration class.

	Attributes
	----------
	device: str
		device path
	power_pin: int
		power pin
	pgm_pin: int
		programming mode pin
	bus_pin: int
		bus enable pin
	baud_rate: int
		uart baud rate
	uart_reset: bool
		reset uart on boot
	i2c_pin: int
		i2c interface pin
	spi_pin: int
		spi interface pin
	uart_pin: int
		uart interface pin

	Methods
	-------
	to_daemon() -> dict:
		Serialize uart configuration into dictionary.
	"""

	def __init__(self, device: str, power_pin: int, pgm_pin: int, bus_pin: int, baud_rate: int,
				uart_reset: bool, i2c_pin: int = None, spi_pin: int = None, uart_pin: int = None) -> None:
		"""
		Construct attributes for uart configuration object.

		Parameters
		----------
		device: str
			device path
		power_pin: int
			power pin
		pgm_pin: int
			programming mode pin
		bus_pin: int
			bus enable pin
		baud_rate: int
			uart baud rate
		uart_reset: bool
			reset uart on boot
		i2c_pin: int
			i2c interface pin
		spi_pin: int
			spi interface pin
		uart_pin: int
			uart interface pin
		"""
		self.device: str = device
		self.power_pin: int = power_pin
		self.pgm_pin: int = pgm_pin
		self.bus_pin: int = bus_pin
		self.baud_rate: int = baud_rate
		self.uart_reset: int = uart_reset
		self.i2c_pin: int = i2c_pin
		self.spi_pin: int = spi_pin
		self.uart_pin: int = uart_pin

	def to_daemon(self) -> dict:
		"""
		Serialize UART configuration into daemon compliant dictionary.

		Parameters
		----------
		None

		Returns
		-------
		dict
			UART configuration serialized into dictionary.
		"""
		config = {
			'component': 'iqrf::IqrfUart',
			'instance': 'iqrf::IqrfUart-' + self.device,
			'IqrfInterface': self.device,
			'powerEnableGpioPin': self.power_pin,
			'busEnableGpioPin': self.bus_pin,
			'pgmSwitchGpioPin': self.pgm_pin,
			'baudRate': self.baud_rate,
			'uartReset': self.uart_reset
		}
		if self.i2c_pin is not None:
			config['i2cEnableGpioPin'] = self.i2c_pin
		if self.spi_pin is not None:
			config['spiEnableGpioPin'] = self.spi_pin
		if self.uart_pin is not None:
			config['uartEnableGpioPin'] = self.uart_pin
		return config
