"""
Copyright 2021-2022 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import os
import time
from distutils.file_util import copy_file
from distutils.dir_util import copy_tree
from distutils.errors import DistutilsFileError
from logging import Logger
from typing import List
import bcrypt
from Setter.Common.ServiceManager import ServiceManager
from Setter.DaemonConfig import DaemonConfig
from Setter.Exceptions.ServiceErrorException import ServiceErrorException
from Setter.NetworkConfig import NetworkConfig
from Setter.WebsocketClient import WebsocketClient
from Setter.Common.Command import Command
from Setter.Common.DaemonApi import DaemonApi
from Setter.Common.GatewayFile import GatewayFile
from Setter.Exceptions.CommandException import CommandException
from Setter.Exceptions.IOErrorException import IOErrorException
from Setter.Logging import Logging
from Setter.Exceptions.SystemErrorException import SystemErrorException


class GatewayConfig:
	"""
	IQRF Gateway device configuration class.

	Methods
	-------
	configure(config: dict) -> None:
		Configure gateway for product utilizing gateway information file.
	generate_gwid() -> None:
		Generate gateway id from mac address and module id.
	generate_token(image: str) -> None:
		Generate pixla token from gateway id.
	configure_tr() -> None:
		Configure module TR configuration.
	configure_system() -> None:
		Configure system time, hostname etc.
	generate_certificate() -> None:
		Generate certificate for daemon and webapp tls.
	configure_grafana() -> None:
		Configure grafana files, set correct permissions and dashboards.
	configure_nodered() -> None:
		Generate webapp API key and configure nodered.
	configure_access() -> None:
		Generate nodered access password and update configuration.
	backup_configuration() -> None:
		Backup configuration of gateway software and network.
	restart_services() -> None:
		Restart systemd services.
	create_boot_flag() -> None:
		Create flag preventing repeated configuration.
	check_image_version() -> None:
		Check and update image version in gateway file.
	"""

	# pylint: disable=unsubscriptable-object
	__id: str = None
	__host: str = None
	__token_path: str = '/etc/gwman/customer_id'
	__setter_path: str = '/var/cache/iqrf-gateway-setter'
	__services: List[str] = [
		'avahi-daemon',
		'resolvconf',
		'NetworkManager',
		'iqrf-gateway-daemon',
		'nginx',
		'mender-client',
		'mender-connect'
	]
	__iqaros: bool = False
	__logger: Logger = Logging.get_logger()

	@classmethod
	def configure(cls, config: dict) -> None:
		"""
		Configure gateway for product utilizing gateway information file.

		Parameters
		----------
		config: dict
			gateway information file

		Returns
		-------
		None
		"""
		cls.__logger.info('Configuring IQRF Gateway device.')
		WebsocketClient.create_connection('localhost', 1338)

		cls.generate_gwid()
		config['gwId'] = cls.__id.upper()
		config['gwToken'] = cls.generate_token(config['gwImage'])
		config['gwHost'] = config['gwHost'].replace('ffffffffffffffff', cls.__id)
		cls.__host = str(config['gwHost'])
		if cls.__host.find('iqaros') != -1:
			cls.__iqaros = True

		DaemonConfig.configure_splitter(cls.__host[:-6])
		DaemonConfig.configure_ide(config['gwProduct'])
		DaemonConfig.configure_mqtt(cls.__host[-22:-6])
		GatewayFile.write(config)
		cls.configure_tr()

		WebsocketClient.close_connection()

		cls.configure_system()
		cls.generate_certificate()

		if cls.__iqaros:
			cls.configure_grafana()
			cls.configure_nodered()
			cls.configure_access()

		cls.backup_configuration()
		cls.restart_services()
		cls.create_boot_flag()

	@classmethod
	def generate_gwid(cls) -> None:
		"""
		Generate gateway id from mac address and module id.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		mac = NetworkConfig.get_mac('eth0').replace(':', '')
		rsp = DaemonApi.read_os(0)
		mid = rsp['data']['rsp']['result']['mid']
		cls.__id = mac[:6] + format(mid & 0xffff, '04x') + mac[6:]
		cls.__logger.info('Generated GW ID: %s', cls.__id)

	@classmethod
	def generate_token(cls, image: str) -> None:
		"""
		Generate PIXLA token from gateway id.

		Generated PIXLA token is stored in /etc/gwman.

		Parameters
		----------
		image: str
			Gateway image string

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorExceptions
			raised if PIXLA token file doesn't exist
		"""
		if not ServiceManager.exists('gwman-client'):
			cls.__logger.info('PIXLA not installed.')
			return None

		product = image.split('-')[0]
		token = product + '-' + cls.__id[4:7] + cls.__id[12:15] + cls.__id[8:11] + cls.__id[0:3]

		try:
			with open(cls.__token_path, mode='w+', encoding='utf-8') as file:
				file.write(token)
			return token
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error

	@classmethod
	def configure_tr(cls) -> None:
		"""
		Configure module TR configuration.

		Sets TX power and RX filter.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		rsp = DaemonApi.write_tr_config(0, {'txPower': 7, 'rxFilter': 7})
		if rsp['data']['rsp']['restartNeeded']:
			DaemonApi.restart_os(0)

	@classmethod
	def configure_system(cls) -> None:
		"""
		Configure system time, hostname etc.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if command to set timezone failed
		"""
		result = Command.execute('timedatectl set-timezone Europe/Prague', True)
		if result.returncode != 0:
			error_str = f'Failed to set timezone: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

	@classmethod
	def generate_certificate(cls) -> None:
		"""
		Generate certificate for daemon and webapp tls.

		Certificate directories are created if they do not exist.
		Existing certificates are removed and new certificates are generated.
		Newly generated files have their permissions fixed.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if certificate or private key command failed
		"""
		cls.__logger.info('Generating certificates.')
		daemon_cert_path = '/etc/iqrf-gateway-daemon/certs/core/'
		webapp_cert_path = '/etc/iqrf-gateway-webapp/certs/'
		files = ['cert.pem', 'privkey.pem']

		if not os.path.exists(daemon_cert_path):
			os.makedirs(daemon_cert_path, exist_ok=True)
		if not os.path.exists(webapp_cert_path):
			os.makedirs(webapp_cert_path, exist_ok=True)

		# remove old files if they exist
		for file in files:
			if os.path.exists(webapp_cert_path + file):
				os.unlink(webapp_cert_path + file)
			if os.path.exists(daemon_cert_path + file):
				os.remove(daemon_cert_path + file)

		# generate private key
		command = f'openssl ecparam -name secp384r1 -genkey -param_enc named_curve -out {daemon_cert_path}{files[1]}'
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to generate private key: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		# generate certificate
		command = 'openssl req -new -x509 -sha256 -nodes -days 3650 ' \
			'-subj "/CN=IQRF Gateway/C=CZ/ST=Hradec Kralove Region/L=Jicin/O=IQRF Tech s.r.o." ' \
			'-key {dir}{keyfile} -out {dir}{certfile}'.format(dir=daemon_cert_path, keyfile=files[1], certfile=files[0])
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to generate certificate: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		# add symlinks to webapp
		for file in files:
			os.chmod(daemon_cert_path + file, 0o600)
			os.symlink(daemon_cert_path + file, webapp_cert_path + file)

	@classmethod
	def configure_grafana(cls) -> None:
		"""
		Configure grafana files, set correct permissions and dashboards.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if permission or owner change command fails
		SystemErrorException
			raised if file manipulation fails
		"""
		result = Command.execute('chown -R root:grafana /etc/grafana')
		if result.returncode != 0:
			error_str = f'Failed to change owner of grafana files: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		result = Command.execute('chown -R grafana:grafana /var/lib/grafana')
		if result.returncode != 0:
			error_str = f'Failed to change owner of grafana lib files: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		try:
			os.remove('/etc/grafana/provisioning/dashboards/sample.yaml')
		except OSError as os_error:
			error_str = f'Failed to remove sample dashboard: {os_error.strerror}.'
			cls.__logger.error(error_str)
			raise SystemErrorException(error_str) from os_error

		try:
			os.chmod('/etc/grafana/grafana.ini', 0o640)
			os.chmod('/etc/grafana/provisioning/dashboards/iqaros.yaml', 0o660)
			os.chmod('/var/lib/grafana/grafana.db', 0o640)
		except OSError as os_error:
			error_str = f'Failed to change permissions of grafana file {os_error.filename}: {os_error.strerror}.'
			cls.__logger.error(error_str)
			raise SystemErrorException(error_str) from os_error

	@classmethod
	def configure_nodered(cls) -> None:
		"""
		Generate webapp API key and configure nodered.

		Generated API key is stored in /etc/iqrf-gateway-webapp.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if command to generate webapp API key fails
		IOErrorException
			raised if API key cannot be saved to file
		"""
		command = 'su - www-data -s /bin/bash -c "iqrf-gateway-webapp-manager api-key:add -n --no-formatting -d nodered"'
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to generate API key for Node-RED: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)
		api_key = result.stdout.strip()

		try:
			with open('/etc/iqrf-gateway-webapp/node-red-token', mode='w', encoding='utf-8') as file:
				file.write(api_key)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error

	@classmethod
	def configure_access(cls) -> None:
		"""
		Generate nodered access password and update configuration.

		Create new webapp, grafana, and node-red flow access password.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if a command generating or configuring access password fails
		IOErrorException
			raised if node-red settings file does not exist or cannot be written to
		"""
		rsp = DaemonApi.read_eeprom(0, 48, 16)
		pdata = rsp['data']['rsp']['result']['pData']
		if 0 in pdata:
			return
		password = ''.join(map(chr, pdata))

		command = f'su - www-data -s /bin/bash -c "iqrf-gateway-webapp-manager user:add -u admin -p {password} -r normal -l en"'
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to create webapp user: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		command = f'grafana-cli admin reset-admin-password {password}'
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to change grafana password: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		try:
			with open('/root/.node-red/settings.js', mode='r', encoding='utf-8') as file:
				settings = file.read()

			nodered_pw = bcrypt.hashpw(password.encode(), bcrypt.gensalt(8)).decode()
			settings.replace('$2b$08$GL\\/ytfjH2scn7AbDwQN\\/D.IbMEiGbQg5PbgSCCrByyWcEerLC3RYW', nodered_pw, 1)

			with open('/root/.node-red/settings.js', mode='w', encoding='utf-8') as file:
				file.write(settings)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error

	@classmethod
	def backup_configuration(cls) -> None:
		"""
		Backup configuration of gateway software and network.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if an error occurs while copying configuration files for backup
		"""
		cls.__logger.info('Backing up default configuration.')
		# create missing paths
		backup_dir = '/usr/share/iqrf-gateway/default/'
		backup_sw = ['controller', 'daemon', 'uploader']

		if ServiceManager.exists('NetworkManager'):
			backup_sw.append('nm')

		# prepare directories
		if not os.path.isdir(backup_dir):
			os.makedirs(backup_dir)
		for software in backup_sw:
			directory = backup_dir + software
			if not os.path.isdir(directory):
				os.makedirs(directory)

		# backup files
		try:
			# backup controller
			copy_tree('/etc/iqrf-gateway-controller/', backup_dir + backup_sw[0], update=1)
			# backup daemon
			copy_tree('/etc/iqrf-gateway-daemon/', backup_dir + backup_sw[1], update=1)
			copy_tree('/var/cache/iqrf-gateway-daemon/metaData', backup_dir + backup_sw[1] + '/metaData', update=1)
			copy_tree('/var/cache/iqrf-gateway-daemon/scheduler', backup_dir + backup_sw[1] + '/scheduler', update=1)
			# backup uploader
			copy_tree('/etc/iqrf-gateway-uploader/', backup_dir + backup_sw[2], update=1)
			# backup network manager
			if ServiceManager.exists('NetworkManager'):
				copy_file('/etc/NetworkManager/NetworkManager.conf', backup_dir + backup_sw[3], update=1)
				copy_tree('/etc/NetworkManager/system-connections', backup_dir + backup_sw[3] + '/system-connections', update=1)
		except DistutilsFileError as dist_error:
			cls.__logger.error(str(dist_error))
			raise IOErrorException(str(dist_error)) from dist_error

	@classmethod
	def restart_services(cls) -> None:
		"""
		Restart systemd services.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		cls.__logger.info('Restarting systemd services.')
		for service in cls.__services:
			try:
				if ServiceManager.is_running(service):
					ServiceManager.restart(service)
			except ServiceErrorException:
				pass
			time.sleep(1)

	@classmethod
	def create_boot_flag(cls) -> None:
		"""
		Create flag preventing repeated configuration.

		Presence of flag file is checked before gateway configuration is executed.
		Gateway is configured if flag file does not exist or if utility is invoked with --force options.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		directory = os.path.dirname(cls.__setter_path)
		if not os.path.exists(directory):
			os.makedirs(directory, exist_ok=True)
		with open(cls.__setter_path, mode='a', encoding='utf-8'):
			os.utime(cls.__setter_path, None)

	@classmethod
	def check_image_version(cls, config: dict) -> None:
		"""
		Check and update image version in gateway file.

		Compares image version in system information and gateway file.
		If the versions mismatch, gateway file is updated according to system information.

		Parameters
		----------
		config: dict
			gateway information file

		Returns
		-------
		None
		"""
		try:
			found = False
			with open('/etc/os-release', mode='r', encoding='utf-8') as file:
				for line in file:
					if not re.search('^VERSION_ID=.*$', line):
						continue
					found = True
					system_version = 'v' + line.split('=')[1].strip().replace('"', '')
					break
			if not found:
				cls.__logger.warning('Could not find OS version entry.')
				return
			pattern = '(?P<product>iqaros|iqube)-(?P<os>armbian|yocto)-(?P<version>v\\d+\\.\\d+\\.\\d+)(-(alpha|beta|rc\\d+))?'
			matches = re.search(pattern, config['gwImage'])
			if matches is None:
				cls.__logger.warning('Failed to match image version.')
				return
			version = matches.group('version')
			if version == system_version:
				return
			cls.__logger.info('New image version detected, updating gateway file to %s.', system_version)
			tokens = [matches.group('product'), matches.group('os'), system_version]
			config['gwImage'] = '-'.join(tokens)
			GatewayFile.write(config)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
